package org.expeditee.items.widgets;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Component;
import java.awt.Font;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.lang.reflect.InvocationTargetException;

import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JTextField;
import javax.swing.JToolBar;
import javax.swing.SwingConstants;
import javax.swing.text.BadLocationException;

import org.expeditee.gui.DisplayIO;
import org.expeditee.gui.FrameMouseActions;
import org.expeditee.gui.FreeItems;
import org.expeditee.gui.MessageBay;
import org.expeditee.io.WebParser;
import org.expeditee.items.Item;
import org.expeditee.items.Text;
import org.expeditee.reflection.JavaFX;
import org.expeditee.settings.network.NetworkSettings;

/**
 * Web browser using a JavaFX WebView.
 * 
 * @author ngw8
 * @author jts21
 */
public class JfxBrowser extends DataFrameWidget {
	
	private static final String BACK = "back";
	private static final String FORWARD = "forward";
	private static final String REFRESH = "refresh";
	private static final String GETFRAME = "getframe";
	
	public static final int VERT_OFFSET = 50;
	public static final int VERT_CROP = 88;
	public static final int HORZ_CROP = 15;
	
	private WebBrowserPanel _browser;
	
	/**
	 * ID of the mouse button that is currently pressed. Uses {@link MouseEvent} constants
	 */
	private int buttonDownId;

	private static class WebBrowserPanel extends JPanel implements ComponentListener {

		private static final long serialVersionUID = 1L;

		Object webview;
		Object jfxPanel;
		JfxBrowser owner;

		/**
		 * ID of the mouse button that is currently pressed.
		 * <p>
		 * Uses {@link MouseEvent} constants
		 */
		private int buttonDownId;

		private JTextField urlField;

		/**
		 * @return A JPanel that either contains a JavaFX Webview or an empty panel if the JFX panel can't be initialized
		 */
		public WebBrowserPanel(final String url) {

			super(new BorderLayout());
			this.addComponentListener(this);
			try {
				jfxPanel = JavaFX.JFXPanel.newInstance();

				// Toolbar (that can't be dragged) to hold the nav buttons, url bar, etc
				JToolBar toolBar = new JToolBar();
				toolBar.setFloatable(false);


				JButton backButton = new JButton("Back");
				toolBar.add(backButton);

				JButton forwardButton = new JButton("Forward");
				toolBar.add(forwardButton);

				urlField = new JTextField(url);
				toolBar.add(urlField);

				JButton convertButton = new JButton("Convert");
				toolBar.add(convertButton);

				backButton.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent arg0) {
						owner.navigateBack();
					}
				});

				forwardButton.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent arg0) {
						owner.navigateForward();
					}
				});
				
				convertButton.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent arg0) {
						owner.getFrame();
					}
				});

				urlField.addActionListener(new ActionListener() {
					@Override
					public void actionPerformed(ActionEvent e) {
						owner.navigate(e.getActionCommand());
					}
				});

				urlField.addKeyListener(new KeyListener() {

					@Override
					public void keyTyped(KeyEvent arg0) {
						// Hiding the cursor when typing, to be more Expeditee-like
						DisplayIO.setCursor(org.expeditee.items.Item.HIDDEN_CURSOR);
					}

					@Override
					public void keyReleased(KeyEvent arg0) { // TODO: remove if not used
					}

					@Override
					public void keyPressed(KeyEvent arg0) { // TODO: remove if not used
					}
				});

				urlField.addMouseMotionListener(new MouseMotionListener() {

					@Override
					public void mouseMoved(MouseEvent arg0) {
						// Checking if the user has been typing - if so, move the cursor to the caret position
						if (DisplayIO.getCursor() == Item.HIDDEN_CURSOR) {
							DisplayIO.setCursor(org.expeditee.items.Item.TEXT_CURSOR);
							try {
								DisplayIO.setCursorPosition(
										// Expeditee gives coords relative to the window, Swing gives it relative to the
										// screen
										urlField.getLocationOnScreen().x + urlField.modelToView(urlField.getCaretPosition() + 1).x - org.expeditee.gui.Browser._theBrowser.getOrigin().x,
										urlField.getLocationOnScreen().y + urlField.modelToView(urlField.getCaretPosition()).y - org.expeditee.gui.Browser._theBrowser.getOrigin().y);

							} catch (BadLocationException e) {}
						} else {
							// Otherwise, move the caret to the cursor location
							urlField.setCaretPosition(urlField.viewToModel(arg0.getPoint()));
						}
					}

					@Override
					public void mouseDragged(MouseEvent arg0) {
						// Selecting text when the mouse is dragged with the middle or right button pressed. If-else
						// determines which direction the user is selecting in
						if (buttonDownId == MouseEvent.BUTTON2 || buttonDownId == MouseEvent.BUTTON3) {
							if (urlField.viewToModel(arg0.getPoint()) < urlField.getSelectionEnd()) {
								urlField.select(urlField.viewToModel(arg0.getPoint()), urlField.getSelectionEnd());
							} else {
								urlField.select(urlField.getSelectionStart(), urlField.viewToModel(arg0.getPoint()));
							}
						}
					}
				});
				
				urlField.addFocusListener(new FocusListener() {
					
					@Override
					public void focusGained(FocusEvent arg0) {
						DisplayIO.setCursor(org.expeditee.items.Item.TEXT_CURSOR);
					}

					@Override
					public void focusLost(FocusEvent arg0) {
						// Restoring the standard cursor, since it is changed to a text cursor when focus is gained
						DisplayIO.setCursor(org.expeditee.items.Item.DEFAULT_CURSOR);
					}
				});

				urlField.addMouseListener(new MouseListener() {

					@Override
					public void mouseReleased(MouseEvent arg0) {
						buttonDownId = MouseEvent.NOBUTTON;

						Text item;

						// If nothing is selected, then select all the text so that it will be copied/moved
						if (urlField.getSelectedText() == null) {
							urlField.selectAll();
						}

						if (arg0.getButton() == MouseEvent.BUTTON3) {
							// Right mouse button released, so copy the selection (i.e. don't remove the original)
							item = DisplayIO.getCurrentFrame().createNewText(urlField.getSelectedText());
							FrameMouseActions.pickup(item);
						} else if (arg0.getButton() == MouseEvent.BUTTON2) {
							// Middle mouse button released, so copy the selection then remove it from the URL field
							item = DisplayIO.getCurrentFrame().createNewText(urlField.getSelectedText());
							urlField.setText(urlField.getText().substring(0, urlField.getSelectionStart()) + urlField.getText().substring(urlField.getSelectionEnd(), urlField.getText().length()));

							FrameMouseActions.pickup(item);
						}
					}

					@Override
					public void mousePressed(MouseEvent arg0) {
						buttonDownId = arg0.getButton();
						
					}

					@Override
					public void mouseExited(MouseEvent arg0) {
					}

					@Override
					public void mouseEntered(MouseEvent arg0) {
					}

					@Override
					public void mouseClicked(MouseEvent arg0) {
					}
				});

				add(toolBar, BorderLayout.PAGE_START);

				this.add((Component) jfxPanel);

				JavaFX.PlatformRunLater.invoke(null, new Runnable() {
					@Override
					public void run() {
						initFx(url);
					}
				});
			} catch (Exception e) {
				e.printStackTrace();
				// if this occurs JavaFX was probably missing
				this.removeAll();
				JLabel label = new JLabel("<html><p style=\"text-align: center;\">Failed to load<br/>Most likely JavaFX could not be found<br/>Make sure you are running at least Java 7<br/>(Preferably Java 8)</p></html>");
				label.setHorizontalAlignment(SwingConstants.CENTER);
				label.setForeground(Color.RED);
				label.setFont(new Font("sans-serif", 0, 20));
				this.add(label);
			}
			
			// Setting bg color to be the same as bg color of widgets when they're being dragged/resized
			this.setBackground(new Color(100, 100, 100));
		}

		/**
		 * Sets up the browser frame. JFX requires this to be run on a new thread.
		 * 
		 * @param url
		 *            The URL to be loaded when the browser is created
		 */
		private void initFx(String url) {
			try {
				this.webview = JavaFX.WebView.newInstance();

				Object scene = JavaFX.SceneConstructor.newInstance(this.webview);

				JavaFX.JFXPanelSetScene.invoke(jfxPanel, scene);

				JavaFX.ReadOnlyObjectPropertyAddListener.invoke(JavaFX.WorkerStateProperty.invoke(JavaFX.WebEngineGetLoadWorker
				        .invoke(JavaFX.WebViewGetEngine.invoke(this.webview))), java.lang.reflect.Proxy.newProxyInstance(
				        		JavaFX.ChangeListener.getClassLoader(), new java.lang.Class[] { JavaFX.ChangeListener },
				        new java.lang.reflect.InvocationHandler() {
					        @Override
					        public Object invoke(Object proxy, java.lang.reflect.Method method, Object[] args)
					                throws java.lang.Throwable {
						        String method_name = method.getName();
						        // Class<?>[] classes = method.getParameterTypes();
						        // public void changed(ObservableValue ov, State oldState, State newState)
						        if (method_name.equals("changed")) {
							        // changed takes 3 args
							        if (args == null || args.length != 3) {
								        return null;
							        }
							        // args[0] is the ObservableValue
							        System.out.println(args[0]);
							        System.out.println(args[0].getClass());
							        // args[2] is the new State
							        if (args[2].getClass() == JavaFX.State) {
								        int id = JavaFX.StateConstants.indexOf(args[2]);
								        switch (id) {
										case 0: // READY
											MessageBay.displayMessage("WebEngine ready");
											break;
										case 1: // SCHEDULED
											MessageBay.displayMessage("Scheduled page load");
											break;
										case 2: // RUNNING
											MessageBay.displayMessage("WebEngine running");

											// Updating the URL bar to display the URL of the page being loaded
											WebBrowserPanel.this.urlField.setText((String) JavaFX.WebEngineGetLocation.invoke(JavaFX.WebViewGetEngine.invoke(WebBrowserPanel.this.webview)));
											break;
										case 3: // SUCCEEDED
											MessageBay.displayMessage("Finished loading page");
											break;
										case 4: // CANCELLED
											MessageBay.displayMessage("Cancelled loading page");
											break;
										case 5: // FAILED
											MessageBay.displayMessage("Failed to load page");
											break;
										}
									}
									System.out.println("\n");
								}
								return null;
							}
				        }));
				
				// Captures mouse click events on webview to enable expeditee like behavior for JavaFX browser.
				JavaFX.NodeSetOnMouseClicked.invoke(this.webview, java.lang.reflect.Proxy.newProxyInstance(
						JavaFX.EventHandler.getClassLoader(), new java.lang.Class[] { JavaFX.EventHandler },
				        new java.lang.reflect.InvocationHandler() {
					        @Override
					        public Object invoke(Object proxy, java.lang.reflect.Method method, Object[] args)
					                throws java.lang.Throwable {
						        
					        	if(method.getName().equals("handle")) {
					        		// Gets button clicked enum values
					        		Object[] buttonEnum = JavaFX.MouseButton.getEnumConstants();
					        		
					        		if(JavaFX.MouseEventGetButton.invoke(args[0]).equals(buttonEnum[1])) /* Left Mouse Button */ {
					        			
					        		} else if(JavaFX.MouseEventGetButton.invoke(args[0]).equals(buttonEnum[2])) /* Middle Mouse Button */ {
					        			
					        		} else if(JavaFX.MouseEventGetButton.invoke(args[0]).equals(buttonEnum[3])) /* Right Mouse Button */ {
					        			// Gets text currently selected in webview
					        			String selection = (String) JavaFX.WebEngineExecuteScript.invoke(JavaFX.WebViewGetEngine.invoke(webview), "window.getSelection().toString()");

					        			// Do nothing if no text is selected
					        			if(!selection.equals("")) {
					        				// Copy text and attach to cursor
											Text t = new Text(selection);
											t.setParent(DisplayIO.getCurrentFrame());
											t.setXY(FrameMouseActions.getX(), FrameMouseActions.getY());
											FrameMouseActions.pickup(t);
					        			} else { System.out.println("no text selected"); }
					        		}
					        	}
					        	
								return null;
							}
				        }));

				// Disables JavaFX webview's right click menu
				JavaFX.WebViewSetContextMenuEnabled.invoke(this.webview, false);
				
				// webEngine.getLoadWorker().stateProperty().addListener(
				// new ChangeListener<State>() {
				// public void changed(ObservableValue ov, State oldState, State newState) {
				// if (newState == State.SUCCEEDED) {
				// stage.setTitle(webEngine.getLocation());
				// }
				// }
				// });

				JavaFX.WebEngineLoad.invoke(JavaFX.WebViewGetEngine.invoke(this.webview), url);
			} catch (Exception e) {
				e.printStackTrace();
			}
		}

		@Override
		public void componentHidden(ComponentEvent e) {
		}

		@Override
		public void componentMoved(ComponentEvent e) {
		}

		@Override
		public void componentResized(ComponentEvent e) {
			// BorderLayout on the JPanel removes the need for explicitly resizing anything here
		}

		@Override
		public void componentShown(ComponentEvent e) {
		}
	}

	public JfxBrowser(Text source, String[] args) {
		// Initial page is either the page stored in the arguments (if there is one stored) or the homepage
		super(source, new WebBrowserPanel((args != null && args.length > 0) ? args[0] : NetworkSettings.homePage), -1, 500, -1, -1, 300, -1);
		_browser = (WebBrowserPanel) _swingComponent;
		_browser.owner = this;
	}

	public void navigate(String url) {
		final String actualURL;
		String urlLower = url.toLowerCase();
		// check if protocol is missing
		if (!(urlLower.startsWith("http://") || url.startsWith("https://") || urlLower.startsWith("ftp://") || urlLower.startsWith("file://"))) {
			// check if it's a search
			int firstSpace = url.indexOf(" ");
			int firstDot = url.indexOf(".");
			int firstSlash = url.indexOf('/');
			int firstQuestion = url.indexOf('?');
			int firstSQ;
			if(firstSlash == -1) {
				firstSQ = firstQuestion;
			} else if(firstQuestion == -1) {
				firstSQ = firstSlash;
			} else {
				firstSQ = -1;
			}
			if(firstDot <= 0 ||                                        // no '.' or starts with '.' -> search
			        (firstSpace != -1 && firstSpace < firstDot + 1) || // ' ' before '.'            -> search
			        (firstSpace != -1 && firstSpace < firstSQ)) {      // no '/' or '?'             -> search
				// make it a search
				actualURL = NetworkSettings.searchEngine + url;
			} else {
				// add the missing protocol
				actualURL = "http://" + url;
			}
		} else {
			actualURL = url;
		}
		System.out.println(actualURL);
		try {
			JavaFX.PlatformRunLater.invoke(null, new Runnable() {
				@Override
				public void run() {
					try {
						JavaFX.WebEngineLoad.invoke(JavaFX.WebViewGetEngine.invoke(JfxBrowser.this._browser.webview), actualURL);
					} catch (Exception e) {
						e.printStackTrace();
					}
				}
			});
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Navigates JfxBrowser back through history. If end of history reached the user is notified via the MessageBay.
	 * Max size of history is 100 by default.
	 */
	public void navigateBack() {
		try {
			final Object webEngine = JavaFX.WebViewGetEngine.invoke(this._browser.webview);
			try {
				JavaFX.PlatformRunLater.invoke(null, new Runnable() {
					@Override
					public void run() {
						try {
							JavaFX.WebHistoryGo.invoke(JavaFX.WebEngineGetHistory.invoke(webEngine), -1);
							FreeItems.getInstance().clear();
						} catch (InvocationTargetException e) {
							MessageBay.displayMessage("Start of History");
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				});
			} catch (Exception e) {
				e.printStackTrace();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Navigates JfxBrowser forward through history. If end of history reached the user is notified via the MessageBay.
	 * Max size of history is 100 by default.
	 */
	public void navigateForward() {
		try {
			final Object webEngine = JavaFX.WebViewGetEngine.invoke(this._browser.webview);
			try {
				JavaFX.PlatformRunLater.invoke(null, new Runnable() {
					@Override
					public void run() {
						try {
							JavaFX.WebHistoryGo.invoke(JavaFX.WebEngineGetHistory.invoke(webEngine), 1);
							FreeItems.getInstance().clear();
						} catch (InvocationTargetException e) {
							MessageBay.displayMessage("End of History");
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				});
			} catch (Exception e) {
				e.printStackTrace();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Refreshes webview by reloading the page.
	 */
	public void refresh() {
		try {
			final Object webEngine = JavaFX.WebViewGetEngine.invoke(this._browser.webview);
			try {
				JavaFX.PlatformRunLater.invoke(null, new Runnable() {
					@Override
					public void run() {
						try {
							JavaFX.WebEngineReload.invoke(webEngine);
							FreeItems.getInstance().clear();
							MessageBay.displayMessage("Page Refreshing");
						} catch (Exception e) {
							e.printStackTrace();
						}
					}
				});
			} catch (Exception e) {
				e.printStackTrace();
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Traverses DOM an turns elements into expeditee items.
	 */
	public void getFrame() {
		try {
			WebParser.parsePage(JavaFX.WebViewGetEngine.invoke(this._browser.webview), DisplayIO.getCurrentFrame());
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Used to drop text items onto JfxBrowser widget. Does nothing if a text item is not attached to cursor. <br>
	 * "back" -> navigates back a page in browser's session history <br>
	 * "forward" -> navigates forward a page in browser's session history <br>
	 * "refresh" -> reloads current page <br>
	 * "getFrame" -> attempts to parse page into an expeditee frame <br>
	 * url -> all other text is assumed to be a url which browser attempts to navigate to
	 * 
	 * @return Whether a JfxBrowser specific event is run.
	 * 
	 */
	@Override
	public boolean ItemsLeftClickDropped() {
		Text carried = null;
		if ((carried = FreeItems.getTextAttachedToCursor()) == null) {	// fails if no text is attached to cursor.
			return false;
		}
		
		if(carried.getText().toLowerCase().equals(BACK)) {
			navigateBack();
		} else if(carried.getText().toLowerCase().equals(FORWARD)) {
			navigateForward();
		} else if(carried.getText().toLowerCase().equals(REFRESH)) {
			refresh();
		} else if(carried.getText().toLowerCase().equals(GETFRAME)) {
			getFrame();
		} else {
			String text = carried.getText().trim();
			this.navigate(text);
			FreeItems.getInstance().clear();
		}
		
		return true;
	}

	/**
	 * Used to enable expeditee like text-widget interaction for middle mouse clicks. Does nothing if a text item is not attached to cursor.
	 * @return false if a text-widget interaction did not occur, true if a text-widget interaction did occur.
	 */
	@Override
	public boolean ItemsMiddleClickDropped() {
		Text t = null;
		if((t = FreeItems.getTextAttachedToCursor()) == null) {	// fails if no text item is attached to the cursor.
			return false;
		}

		Point p = new Point(FrameMouseActions.getX() - this.getX(), FrameMouseActions.getY() - this.getY());

		if(!_browser.urlField.contains(p)) {					// fails if not clicking on urlField
			return false;
		}

		// Inserts text in text item into urlField at the position of the mouse.
		String s = _browser.urlField.getText();
		int index = _browser.urlField.viewToModel(new Point(p.x - _browser.urlField.getX(), p.y - _browser.urlField.getY()));
		s = s.substring(0, index) + t.getText() + s.substring(index);
		_browser.urlField.setText(s);

		FreeItems.getInstance().clear();						// removed held text item - like normal expeditee middle click behaviour.

		return true;
	}
	
	/**
	 * Used to enable expeditee like text-widget interaction for right mouse clicks. Does nothing if a text item is not attached to cursor.
	 * @return false if a text-widget interaction did not occur, true if a text-widget interaction did occur.
	 */
	@Override
	public boolean ItemsRightClickDropped() {
		Text t = null;
		if((t = FreeItems.getTextAttachedToCursor()) == null) {	// fails if no text item is attached to the cursor.
			return false;
		}

		Point p = new Point(FrameMouseActions.getX() - this.getX(), FrameMouseActions.getY() - this.getY());

		if(!_browser.urlField.contains(p)) {					// fails if not clicking on urlField
			return false;
		}

		// Inserts text in text item into urlField at the position of the mouse.
		String s = _browser.urlField.getText();
		int index = _browser.urlField.viewToModel(new Point(p.x - _browser.urlField.getX(), p.y - _browser.urlField.getY()));
		s = s.substring(0, index) + t.getText() + s.substring(index);
		_browser.urlField.setText(s);

		return true;
	}
	
	@Override
	protected String[] getArgs() {
		String[] r = null;
		if (this._browser.webview != null) {
			try {
				r = new String[] { (String) JavaFX.WebEngineGetLocation.invoke(JavaFX.WebViewGetEngine.invoke(this._browser.webview)) };
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return r;
	}
}
